package kc.mega.dodge;

import kc.mega.learning.GFBins;
import kc.mega.learning.KNNHitModel;
import kc.mega.learning.Model;

public class DangerModels {
  public static Model simple(GFBins bins) {
    return new KNNHitModel.Builder("Simple")
        .bins(bins)
        .features(new String[] {"bft", "vel", "latVel", "advDir", "maeWallAhead"})
        .params(new double[][] {{0.5}, {2}, {6}, {1}, {2}})
        .distanceScale(-2)
        .neighborhoodSizeDivider(2)
        .maxNeighbors(30)
        .maxTreeSize(3000)
        .build();
  }

  public static Model simple2(GFBins bins) {
    return new KNNHitModel.Builder("Simple2")
        .bins(bins)
        .features(new String[] {"bft", "accel", "vel", "advDir", "maeWallAhead", "stickWallAhead"})
        .params(new double[][] {{1}, {1}, {2}, {1}, {1}, {1}})
        .distanceScale(-0.5)
        .neighborhoodSizeDivider(2)
        .maxNeighbors(30)
        .maxTreeSize(3000)
        .build();
  }

  public static Model antiThorn(GFBins bins) {
    return new KNNHitModel.Builder("Thorn")
      .bins(bins)
      .features(new String[] {"bft", "accel", "vel", "orbitalWallAhead", "orbitalWallReverse", "maeWallAhead", "maeWallReverse", "shotsFired"})
      .params(new double[][] {{10.3438}, {1.6888}, {2.1116}, {1.7250}, {0.9364}, {1.4386}, {2.1662}, {3.1504}})
      .distanceScale(-1.6325)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiSedan(GFBins bins) {
    return new KNNHitModel.Builder("Sedan")
      .bins(bins)
      .features(new String[] {"bft", "accel", "latVel", "maeWallAhead", "shotsFired"})
      .params(new double[][] {{7.8667}, {0.7579}, {1.3128}, {1.1977}, {3.9877}})
      .distanceScale(-2.4324)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiChiva(GFBins bins) {
    return new KNNHitModel.Builder("Chiva")
      .bins(bins)
      .features(new String[] {"bft", "accel", "latVel", "vChangeTimer", "maeWallAhead", "maeWallReverse", "shotsFired"})
      .params(new double[][] {{7.7335}, {1.1874}, {2.2441}, {2.3042}, {0.9538}, {0.4040}, {2.8453}})
      .distanceScale(-1.9306)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiDruss(GFBins bins) {
    return new KNNHitModel.Builder("Druss")
      .bins(bins)
      .features(new String[] {"bft", "accel", "vel", "latVel", "advDir", "vChangeTimer", "decelTimer", "distanceLast10", "maeWallAhead", "maeWallReverse", "stickWallAhead", "stickWallReverse", "shotsFired"})
      .params(new double[][] {{7.4265}, {3.2607}, {0.3331}, {3.0054}, {4.5670}, {0.6459}, {2.3745}, {2.8849}, {4.1025}, {1.2191}, {0.1208}, {0.3755}, {0.5554}})
      .distanceScale(-0.2199)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiDiamond(GFBins bins) {
    return new KNNHitModel.Builder("Diamond")
      .bins(bins)
      .features(new String[] {"bft", "accel", "latVel", "advDir", "vChangeTimer", "orbitalWallAhead", "orbitalWallReverse", "maeWallAhead", "maeWallReverse", "shotsFired"})
      .params(new double[][] {{10.4572}, {1.5877}, {0.2986}, {6.1020}, {3.0055}, {2.1124}, {2.6054}, {2.5548}, {2.0901}, {0.3814}})
      .distanceScale(-0.4383)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiWaveSerpent(GFBins bins) {
    return new KNNHitModel.Builder("Waveserpent")
      .bins(bins)
      .features(new String[] {"bft", "accel", "latVel", "dirChangeTimer", "decelTimer", "distanceLast10", "distanceLast20", "orbitalWallAhead", "orbitalWallReverse", "maeWallAhead", "stickWallAhead", "stickWallReverse", "shotsFired"})
      .params(new double[][] {{8.7311}, {2.5460}, {1.7778}, {0.7067}, {1.5468}, {1.0153}, {0.9312}, {2.5778}, {2.1063}, {2.1637}, {2.3050}, {0.2110}, {3.1157}})
      .distanceScale(-0.2525)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiCassiusClay(GFBins bins) {
    return new KNNHitModel.Builder("Cassiusclay")
      .bins(bins)
      .features(new String[] {"bft", "accel", "latVel", "advDir", "vChangeTimer", "decelTimer", "distanceLast20", "orbitalWallAhead", "orbitalWallReverse", "maeWallAhead", "stickWallAhead", "stickWallReverse", "shotsFired"})
      .params(new double[][] {{12.0932}, {1.8222}, {2.3525}, {2.7026}, {0.6014}, {1.3150}, {0.9451}, {2.0092}, {1.3457}, {2.6058}, {0.5738}, {0.5877}, {3.0893}})
      .distanceScale(-0.3222)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiKomarious(GFBins bins) {
    return new KNNHitModel.Builder("Komarious")
      .bins(bins)
      .features(new String[] {"bft", "accel", "latVel", "advDir", "vChangeTimer", "orbitalWallAhead", "orbitalWallReverse", "stickWallAhead", "stickWallReverse", "shotsFired"})
      .params(new double[][] {{2.2555}, {1.6461}, {2.4056}, {2.4558}, {2.4851}, {2.9594}, {1.4271}, {1.6963}, {1.8807}, {4.0817}})
      .distanceScale(-1.0133)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiSplinter(GFBins bins) {
    return new KNNHitModel.Builder("Splinter")
      .bins(bins)
      .features(new String[] {"bft", "latVel", "maeWallAhead"})
      .params(new double[][] {{5.5207}, {2.1004}, {0.9032}})
      .distanceScale(-2.7087)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiRaikoNano(GFBins bins) {
    return new KNNHitModel.Builder("RaikoNano")
      .bins(bins)
      .features(new String[] {"bft", "vel", "advDir", "maeWallAhead"})
      .params(new double[][] {{2.3633}, {3.7156}, {0.3229}, {0.5192}})
      .distanceScale(-1.5953)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiWaveShark(GFBins bins) {
    return new KNNHitModel.Builder("Waveshark")
      .bins(bins)
      .features(new String[] {"bft", "accel", "latVel", "advDir", "dirChangeTimer", "decelTimer", "distanceLast20", "orbitalWallAhead", "orbitalWallReverse", "maeWallAhead", "maeWallReverse"})
      .params(new double[][] {{4.7128}, {2.9091}, {3.6226}, {1.7127}, {1.6968}, {3.5238}, {0.9638}, {1.0648}, {0.7448}, {2.4531}, {1.1200}})
      .distanceScale(-0.1762)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiWaylander(GFBins bins) {
    return new KNNHitModel.Builder("Waylander")
      .bins(bins)
      .features(new String[] {"accel", "vel", "latVel", "vChangeTimer", "orbitalWallAhead", "stickWallAhead"})
      .params(new double[][] {{5.0547}, {3.5802}, {2.0354}, {1.1943}, {0.8919}, {0.9510}})
      .distanceScale(-0.1188)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiMicroAspid(GFBins bins) {
    return new KNNHitModel.Builder("Microaspid")
      .bins(bins)
      .features(new String[] {"bft", "accel", "latVel", "vChangeTimer", "decelTimer", "orbitalWallAhead", "orbitalWallReverse", "stickWallAhead", "stickWallReverse"})
      .params(new double[][] {{7.0838}, {0.0026}, {5.4960}, {1.7377}, {1.1942}, {0.4861}, {0.4431}, {2.5026}, {1.7946}})
      .distanceScale(-0.1865)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  public static Model antiGrinnik(GFBins bins) {
    return new KNNHitModel.Builder("Grinnik")
      .bins(bins)
      .features(new String[] {"bft", "accel", "vel", "vChangeTimer", "shotsFired"})
      .params(new double[][] {{11.8852}, {1.0335}, {2.3859}, {1.8833}, {3.6499}})
      .distanceScale(-2.1291)
      .neighborhoodSizeDivider(2)
      .maxNeighbors(30)
      .maxTreeSize(3000)
      .build();
  }

  // diamond-druss
  public static Model flattener1(GFBins bins) {
    return new KNNHitModel.Builder("Flattener1")
      .bins(bins)
      .features(new String[] {"virtuality", "bft", "accel", "vel", "maeWallAhead", "shotsFired"})
      .params(new double[][] {{2.8498}, {3.2063}, {2.8554}, {4.0240}, {6.0501}, {3.9666}})
      .distanceScale(-0.0956)
      .neighborhoodSizeDivider(5)
      .maxNeighbors(50)
      .maxTreeSize(5000)
      .build();
  }

  // waveserpent-cassiusclay-gressuffard
  public static Model flattener2(GFBins bins) {
    return new KNNHitModel.Builder("Flattener2")
      .bins(bins)
      .features(new String[] {"virtuality", "bft", "accel", "vel", "advDir", "maeWallAhead", "maeWallReverse", "stickWallAhead", "shotsFired"})
      .params(new double[][] {{2.7282}, {2.8465}, {1.3839}, {5.0453}, {3.4051}, {2.1750}, {1.2043}, {1.7921}, {5.7242}})
      .distanceScale(-0.1651)
      .neighborhoodSizeDivider(5)
      .maxNeighbors(50)
      .maxTreeSize(5000)
      .build();
  }

  // tomcat-gilgalad
  public static Model flattener3(GFBins bins) {
    return new KNNHitModel.Builder("Flattener3")
      .bins(bins)
      .features(new String[] {"virtuality", "bft", "accel", "vel", "advDir", "dirChangeTimer", "maeWallAhead", "maeWallReverse", "stickWallAhead", "stickWallReverse", "shotsFired"})
      .params(new double[][] {{2.1865}, {4.9589}, {0.8323}, {1.0450}, {2.2741}, {0.6626}, {0.7798}, {0.7008}, {1.3206}, {1.2121}, {2.0052}})
      .distanceScale(-0.6058)
      .neighborhoodSizeDivider(5)
      .maxNeighbors(50)
      .maxTreeSize(5000)
      .build();
  }

   // beepboop-komarious-cunobelin
  public static Model flattener4(GFBins bins) {
    return new KNNHitModel.Builder("Flattener4")
      .bins(bins)
      .features(new String[] {"virtuality", "bft", "accel", "vel", "advDir", "dirChangeTimer", "maeWallAhead", "maeWallReverse", "shotsFired"})
      .params(new double[][] {{2.1167}, {2.3384}, {0.5146}, {1.5453}, {2.2296}, {1.0241}, {2.3923}, {1.1432}, {4.4724}})
      .distanceScale(-0.5052)
      .neighborhoodSizeDivider(5)
      .maxNeighbors(50)
      .maxTreeSize(5000)
      .build();
  }
}
