package kc.mega.misc;

import kc.mega.game.BattleField;
import kc.mega.game.GameState;
import kc.mega.utils.Geom;
import kc.mega.utils.KUtils;

import java.awt.geom.Point2D;

import robocode.AdvancedRobot;

public class Scanner {
  private static final double SCAN_WIDTH = Math.PI / 32;
  private static final int REVERALS_BEFORE_SWEEP = 3;

  private final AdvancedRobot bot;
  private long lastScannedTime;
  private int radarTurnDirection;
  private int radarReversals;

  public Scanner(AdvancedRobot bot) {
    this.bot = bot;
  }

  public void init() {
    lastScannedTime = -1;
    radarReversals = 0;
    radarTurnDirection = KUtils.nonzeroSign(Geom.offset(
        new Point2D.Double(bot.getX(), bot.getY()),
        BattleField.INSTANCE.getCenter(),
        bot.getRadarHeadingRadians()));
  }

  public long getLastScannedTime() {
    return lastScannedTime;
  }

  public void sweep(GameState gs) {
    if (lastScannedTime < 0) {
      bot.setTurnGunRightRadians(radarTurnDirection * Double.POSITIVE_INFINITY);
      bot.setTurnRadarRightRadians(radarTurnDirection * Double.POSITIVE_INFINITY);
    } else if(lastScannedTime < bot.getTime()) {
      if (radarReversals < REVERALS_BEFORE_SWEEP) {
        int newRadarTurnDirection = KUtils.nonzeroSign(
            Geom.offset(gs.getMyState().location, gs.getEnemyState().location,
                bot.getRadarHeadingRadians()));
        if (newRadarTurnDirection != radarTurnDirection) {
          radarTurnDirection = newRadarTurnDirection;
          radarReversals++;
        }
      }
      bot.setTurnRadarRightRadians(radarTurnDirection * Double.POSITIVE_INFINITY);
    }
  }

  public void scan(GameState gs) {
    if (lastScannedTime < 0) {
      bot.setTurnGunRightRadians(-radarTurnDirection * Math.PI / 18);
    }
    radarReversals = 0;
    lastScannedTime = gs.gameTime;
    double radarTurn = Geom.offset(
        gs.getMyState().location, gs.getEnemyState().location, bot.getRadarHeadingRadians());
    bot.setTurnRadarRightRadians(radarTurn + KUtils.nonzeroSign(radarTurn) * SCAN_WIDTH);
  }
}
